-- PWMG Pulse Width Modulation Generator
--
-- pwmg.vhd
--
-- The function of this module is to generate pwm signals used to 
-- control the onboard led 
------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity pwm is
    Port ( sysclk : in std_logic;   -- 12Mhz. clock
           reset_n : in std_logic;
           address_out: out std_logic_vector (7 downto 0);
           rom_data: in std_logic_vector (8 downto 0);
           clk_12M : out std_logic;
           output: out std_logic           
); 
end pwm;

architecture behavioral of pwm is

signal tcount: std_logic_vector (11 downto 0);
signal address: std_logic_vector (7 downto 0);
signal fcount: std_logic_vector (7 downto 0);
signal shadow: std_logic_vector ( 7 downto 0);
signal pwm_cnt: std_logic_vector ( 7 downto 0);
signal disp_val: std_logic_vector ( 7 downto 0);
signal pwm_clk: std_logic;
signal clk: std_logic; --12mhz. clock
signal fpulse: std_logic;
signal fpulse_ff: std_logic;
signal fpulse_rd: std_logic;


begin

address_out <= address;
clk_12M <= clk;

process (sysclk, reset_n)
begin
   if reset_n='0' then
         clk <= '0';
     elsif sysclk='1' and sysclk'event then
           clk <= not clk;
     end if;
end process;





-- Clock Generation circuit, This circuit will create a 64hz
-- frame update pulse and the 16.393 Khz. PWM pulse from a 
-- 12Mhz. clock. Acual counts are as follows:

-- 16.393K = 732d or 2DCh (12Mhz.) ticks 
-- 64 Hz = 187392 or 2DC00h (12Mhz.) ticks

--pwm clock circuit

process (clk, reset_n)
begin
   if reset_n='0' then
         tcount(11 downto 0) <= x"000";
         pwm_clk <= '0';
     elsif clk='1' and clk'event then
        if (tcount = x"16e") then -- half of 2DC because 2DC is entire period          
            tcount <= x"000";
            pwm_clk <= not pwm_clk;
         else
            tcount <= tcount + 1;
            pwm_clk <= pwm_clk;
        end if;
     end if;
end process;

-- Frame pulse circuit
process (pwm_clk, reset_n)
begin
   if reset_n='0' then
         fcount(7 downto 0) <= x"00";
         fpulse <= '0';
     elsif pwm_clk='1' and pwm_clk'event then
        if (fcount = x"ff") then -- FF = 256 which is the number of PWM_clks in a frame          
            fcount <= x"00";
            fpulse <= '1';
         else
            fcount <= fcount + 1;
            fpulse <= '0';
        end if;
     end if;
end process;

-- FPULSE edgedetect, because each FPULSE is 61us wide. We use
-- an edge detect circuit to identify the rising edge, this 
-- allows us to use the faster 12Mhz. clock to read memory and 
-- update the PWM display register. It also means we will only 
-- update once for each FPULSE.
  
process (clk, reset_n)
begin
   if reset_n='0' then
         fpulse_ff <= '0';
     elsif clk='1' and clk'event then
         fpulse_ff <= fpulse;
      end if;
  end process;

process (clk, reset_n)
begin
   if reset_n='0' then
         fpulse_rd <= '0';
     elsif clk='1' and clk'event then
           if (fpulse = '1' and fpulse_ff = '0') then -- rising edge;
                 fpulse_rd <= '1';
        else
                 fpulse_rd <= '0';
      end if;
     end if;
  end process;

-- Memory Fetch circuit, this circuit will fetch an updated 
-- pwm setting from an instantiated memory and present it to 
-- the PWM output circuit. If the rollover flag rom_data(11) is
-- set the address will be reset to 0.
 
process (clk, reset_n)
begin
   if reset_n='0' then
         address(7 downto 0) <= x"00";
         shadow(7 downto 0) <= x"00";
     elsif clk='1' and clk'event then
        if (fpulse_rd = '1') then
 			shadow <= rom_data(7 downto 0);
            address <= address +1;
         elsif (rom_data(8) = '1') then
 			shadow <= shadow;
            address <= x"00";
         else
            shadow <= shadow;
            address <= address;
        end if;
     end if;
end process;

-- PWM display circuit

process (pwm_clk, reset_n)
begin
   if reset_n='0' then
         output <= '0';
         pwm_cnt <= x"01";
         disp_val <= x"00";
     elsif pwm_clk='1' and pwm_clk'event then
        if (pwm_cnt = x"ff") then
             pwm_cnt <= x"00"; 
             disp_val <= shadow;
         elsif (pwm_cnt < disp_val) then
               output <= '0'; -- display is on
               pwm_cnt <= pwm_cnt +1;
               disp_val <= disp_val;
           else 
			    output <= '1'; -- display is off
               pwm_cnt <= pwm_cnt +1;
               disp_val <= disp_val;
        end if;
     end if;
end process;



 

end behavioral;
